<?php

namespace Nest\Http;

use Nest\Application;
use Nest\Http\Exceptions\Renderer\ViewNotFoundException;
use Nyholm\Psr7\Factory\Psr17Factory;
use Nyholm\Psr7Server\ServerRequestCreator;
use Override;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use function Nest\Utils\array_first;
use function Nest\Utils\path_join;

/**
 * HTTP service manager.
 */
class Http
{
	/**
	 * Current request instance.
	 * @var ServerRequestInterface
	 */
	private ServerRequestInterface $request;

	/**
	 * @param Application $application The application.
	 * @param HttpConfiguration $configuration HTTP service configuration.
	 */
	public function __construct(protected Application $application, protected HttpConfiguration $configuration)
	{
	}

	/**
	 * Get HTTP debug status.
	 * @return bool HTTP debug status.
	 */
	public function getDebug(): bool
	{
		return $this->configuration->getDebug();
	}

	/**
	 * Get default views path, if there is one.
	 * @return string|null Default views path.
	 */
	public function getDefaultViewsPath(): ?string
	{
		return $this->configuration->getViewsPath();
	}

	/**
	 * Resolve the given view name to an absolute path to the actual view file.
	 * @param string $viewName Name of the view to resolve.
	 * @return string Absolute path to the view file.
	 * @throws ViewNotFoundException
	 */
	public function resolveView(string $viewName): string
	{
		if (file_exists($viewName))
			// If the view name is already a valid file, just return it.
			return $viewName;
		else
		{ // Try to build a view absolute path with the given view name.
			$viewPath = path_join($this->getDefaultViewsPath(), "$viewName.php");
			if (file_exists($viewPath)) return $viewPath;
			$viewPath = path_join($this->getDefaultViewsPath(), "$viewName");
			if (file_exists($viewPath)) return $viewPath;

			// No view path found, throw an exception.
			throw new ViewNotFoundException();
		}
	}

	/**
	 * Parse the current server request from global context.
	 * @return ServerRequestInterface Parsed server request.
	 */
	protected function parseRequest(): ServerRequestInterface
	{
		// Initialize server request.
		$psr17Factor = new Psr17Factory();
		$this->request = (new ServerRequestCreator(
			$psr17Factor,
			$psr17Factor,
			$psr17Factor,
			$psr17Factor,
		))->fromGlobals();

		if (str_starts_with(array_first($this->request->getHeader("ContentType")), "application/json"))
			// If body is supposed to be JSON-encoded, decode JSON.
			$this->request = $this->request->withParsedBody(json_decode($this->request->getBody()->getContents()));

		return $this->request;
	}

	/**
	 * Build a chain of request handlers (with middlewares).
	 * @return RequestHandlerInterface
	 */
	protected function buildCallChain(): RequestHandlerInterface
	{
		// Set the final request handler as the current one.
		$currentHandler = $this->configuration->getRequestHandler();

		foreach ($this->configuration->getMiddlewares() as $middleware)
		{ // Add each middleware.
			// The current handler is calling the middleware which will call the previously current one if needed.
			$currentHandler = new class($middleware, $currentHandler) implements RequestHandlerInterface {
				public function __construct(private MiddlewareInterface $middleware, private RequestHandlerInterface $requestHandler)
				{}

				/**
				 * @inheritDoc
				 */
				#[Override] public function handle(ServerRequestInterface $request): ResponseInterface
				{
					return $this->middleware->process($request, $this->requestHandler);
				}
			};
		}

		// Return the built request handler.
		return $currentHandler;
	}

	/**
	 * Start HTTP engine.
	 * @return void
	 */
	public function __invoke(): void
	{
		// Create and initialize request object.
		$this->parseRequest();

		// Get response from handler and send it.
		$response = $this->buildCallChain()->handle($this->request);
		$this->output($response);
	}

	/**
	 * Output a given response.
	 * @param ResponseInterface $response The response to output.
	 * @return void
	 */
	public function output(ResponseInterface $response): void
	{
		// Send status code.
		http_response_code($response->getStatusCode());

		// Send headers.
		foreach ($response->getHeaders() as $name => $values)
		{ // Send each header.
			foreach ($values as $value)
			{ // Send each header value.
				header("$name: $value", false);
			}
		}

		// Send body.
		echo $response->getBody()->getContents();
	}

	/**
	 * Get current request instance, if there is one.
	 * @return ServerRequestInterface|null
	 */
	public function getRequest(): ?ServerRequestInterface
	{
		return $this->request ?? null;
	}
}
