<?php

namespace Nest\Http;

use Nest\Exceptions\InvalidTypeException;
use Nest\Services\ServiceConfiguration;
use Override;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use function Nest\Utils\expect_type;

/**
 * HTTP configuration class.
 */
class HttpConfiguration extends ServiceConfiguration
{
	/**
	 * HTTP debug mode.
	 * This is NOT suitable for production as it outputs clearly some code and may share secrets.
	 * @var bool
	 */
	private bool $debug = false;

	/**
	 * HTTP exception renderer class.
	 * @var string
	 */
	private string $exceptionRenderer = ExceptionRenderer::class;

	/**
	 * Default views directory path, if there is one.
	 * @var string|null
	 */
	private ?string $viewsPath = null;

	/**
	 * The main request handler.
	 * @var RequestHandlerInterface
	 */
	private RequestHandlerInterface $requestHandler;

	/**
	 * List of global middlewares.
	 * @var MiddlewareInterface[]
	 */
	private array $middlewares = [];

	public function __construct()
	{
		// Add default middlewares.
		$this->addDefaultMiddlewares();
	}

	/**
	 * Set debug mode.
	 * @param bool $debug Debug mode status.
	 * @return $this
	 */
	public function debug(bool $debug = true): static
	{
		$this->debug = $debug;
		return $this;
	}

	/**
	 * Set exception renderer class.
	 * @param string $exceptionRenderer The exception renderer to use and an HTTP exception is thrown.
	 * @return $this
	 * @throws InvalidTypeException
	 */
	public function exceptionRenderer(string $exceptionRenderer): static
	{
		expect_type($exceptionRenderer, ExceptionRenderer::class);
		$this->exceptionRenderer = $exceptionRenderer;
		return $this;
	}

	/**
	 * Set default views path.
	 * @param string|null $viewsPath Path to views directory. NULL to remove default paths directory.
	 * @return $this
	 */
	public function viewsPath(?string $viewsPath): static
	{
		$this->viewsPath = $viewsPath;
		return $this;
	}

	/**
	 * Set HTTP request handler.
	 * @param (callable(ServerRequestInterface): ResponseInterface)|RequestHandlerInterface $requestHandler
	 * @return $this
	 */
	public function requestHandler(callable|RequestHandlerInterface $requestHandler): static
	{
		if (!($requestHandler instanceof RequestHandlerInterface))
			// Convert a callable to a request handler interface.
			$requestHandler = new class($requestHandler) implements RequestHandlerInterface {
				private $requestHandler;

				public function __construct(callable $requestHandler)
				{
					$this->requestHandler = $requestHandler;
				}

				/**
				 * @inheritDoc
				 */
				#[Override] public function handle(ServerRequestInterface $request): ResponseInterface
				{
					return ($this->requestHandler)($request);
				}
			};

		$this->requestHandler = $requestHandler;
		return $this;
	}

	/**
	 * Reset middlewares.
	 * @return $this
	 */
	public function resetMiddlewares(): static
	{
		$this->middlewares = [];
		return $this;
	}

	/**
	 * Add default middlewares.
	 * They are added in the constructor, so if you didn't reset the middlewares, running this isn't required.
	 * @return $this
	 */
	public function addDefaultMiddlewares(): static
	{
		$this->middlewares[] = new ErrorsMiddleware($this);
		return $this;
	}

	/**
	 * Add a new middleware to the stack.
	 * What is added first is called first in request handling.
	 * @param (callable(ServerRequestInterface, RequestHandlerInterface): ResponseInterface)|MiddlewareInterface $middleware
	 * @return $this
	 */
	public function addMiddleware(callable|MiddlewareInterface $middleware): static
	{
		if (!($middleware instanceof MiddlewareInterface))
			// Convert a callable to a middleware handler interface.
			$middleware = new class($middleware) implements MiddlewareInterface {
				private $middleware;

				public function __construct(callable $middleware)
				{
					$this->middleware = $middleware;
				}

				/**
				 * @inheritDoc
				 */
				#[Override] public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface
				{
					return ($this->middleware)($request, $handler);
				}
			};

		$this->middlewares[] = $middleware;
		return $this;
	}

	/**
	 * Get HTTP debug status.
	 * @return bool HTTP debug status.
	 */
	public function getDebug(): bool
	{
		return $this->debug;
	}

	/**
	 * Get exception renderer instance.
	 * @return string
	 */
	public function getExceptionRenderer(): string
	{
		return $this->exceptionRenderer;
	}

	/**
	 * Get default views path.
	 * @return string|null Default views path, if there is one.
	 */
	public function getViewsPath(): ?string
	{
		return $this->viewsPath;
	}

	/**
	 * Get configured request handler.
	 * @return RequestHandlerInterface
	 */
	public function getRequestHandler(): RequestHandlerInterface
	{
		return $this->requestHandler;
	}

	/**
	 * Get configured middlewares stack.
	 * @return MiddlewareInterface[]
	 */
	public function getMiddlewares(): array
	{
		return $this->middlewares;
	}
}
