<?php

namespace Nest\Http\Router;

use Nest\Application;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;

/**
 * Route definition class.
 */
class Route
{
	/**
	 * List of allowed request methods.
	 * All methods are allowed by default (empty array).
	 * @var string[]
	 */
	protected array $methods = [];

	/**
	 * List of allowed servers names.
	 * All servers names are allowed by default (empty array).
	 * @var string[]
	 */
	protected array $serversNames = [];

	/**
	 * Request handler for defined route (URI parameters as function parameters).
	 * @var callable(mixed...): ResponseInterface
	 */
	protected $handler;

	/**
	 * Create a route from its URI.
	 * @param string $uri Route URI.
	 */
	public function __construct(protected string $uri)
	{}

	/**
	 * Add allowed request methods (GET, POST, etc.).
	 * @param string ...$method Allowed method.
	 * @return $this
	 */
	public function method(string ...$method): static
	{
		array_push($this->methods, ...$method);
		return $this;
	}
	/**
	 * Allow all request methods.
	 * @return $this
	 */
	public function allMethods(): static
	{
		$this->methods = [];
		return $this;
	}
	/**
	 * Allow GET request method.
	 * @return $this
	 */
	public function get(): static
	{
		return $this->method("GET");
	}
	/**
	 * Allow POST request method.
	 * @return $this
	 */
	public function post(): static
	{
		return $this->method("POST");
	}
	/**
	 * Allow PATCH request method.
	 * @return $this
	 */
	public function patch(): static
	{
		return $this->method("PATCH");
	}
	/**
	 * Allow DELETE request method.
	 * @return $this
	 */
	public function delete(): static
	{
		return $this->method("DELETE");
	}

	/**
	 * Reset allowed servers names to all by default.
	 * @return $this
	 */
	public function resetServersNames(): static
	{
		$this->serversNames = [];
		return $this;
	}
	/**
	 * Add an allowed server name for the given route.
	 * @return $this
	 */
	public function serverName(string $serverName): static
	{
		$this->serversNames[] = $serverName;
		return $this;
	}

	/**
	 * Set request handler for given route.
	 * @param callable(mixed...): ResponseInterface $handler Request handler. Will be called with parameters in defined URI.
	 * @return $this
	 */
	public function handler(callable $handler): static
	{
		$this->handler = $handler;
		return $this;
	}

	/**
	 * Set a controller method as handler.
	 * @param string $controllerClass The controller class.
	 * @param string $controllerMethod The method to call in the controller.
	 * @return $this
	 */
	public function controller(string $controllerClass, string $controllerMethod): static
	{
		$this->handler = function (...$args) use ($controllerClass, $controllerMethod) {
			return call_user_func_array([new $controllerClass(Application::get()), $controllerMethod], $args);
		};
		return $this;
	}

	/**
	 * Get route URI.
	 * @return string
	 */
	public function getUri(): string
	{
		return $this->uri;
	}

	/**
	 * Get request handler for given route.
	 * @return callable(mixed...): ResponseInterface Request handler. Will be called with parameters in defined URI.
	 */
	public function getHandler(): callable
	{
		return $this->handler;
	}

	/**
	 * Check if the route matches a request.
	 * @param ServerRequestInterface $request The request to match.
	 * @return bool True if the route matches the request.
	 */
	public function match(ServerRequestInterface $request): bool
	{
		return (empty($this->methods) || in_array($request->getMethod(), $this->methods)) &&
			(empty($this->serversNames) || in_array($request->getUri()->getHost(), $this->serversNames)) &&
			!empty($this->handler);
	}
}
