<?php

namespace Nest\Http\Router;

use Nest\Http\Exceptions\NotFoundException;
use Psr\Http\Message\ServerRequestInterface;

/**
 * Routes definition class.
 */
abstract class Routes
{
	/**
	 * Define routes.
	 * @return void
	 */
	public abstract function define(): void;

	/**
	 * Main routes tree node.
	 * @var RoutesTree
	 */
	protected RoutesTree $tree;

	/**
	 * Create a new routes definition object.
	 */
	public function __construct()
	{
		$this->tree = new RoutesTree("");
	}

	/**
	 * Split the given URI it in route nodes names.
	 * @return string[] Routes nodes names.
	 */
	private function splitUri(string $uri): array
	{
		$uri = trim($uri, "/ \n\r\t\v\0");
		return array_map(fn ($node) => trim($node, "/ \n\r\t\v\0"), explode("/", $uri));
	}

	/**
	 * Add a new route for the given URI.
	 * @param Route $route The route to add.
	 * @return void
	 */
	protected function addRoute(Route $route): void
	{
		$nodes = $this->splitUri($route->getUri());
		// Register the new route with its split URI.
		$this->tree->register($nodes, $route);
	}

	/**
	 * Define a new route.
	 * @param string $uri The route URI.
	 * @return Route The defined route.
	 */
	public function route(string $uri): Route
	{
		// Create a route with the given URI and add it to the list.
		$this->addRoute($route = new Route($uri));

		// Return created route.
		return $route;
	}

	/**
	 * Try to match a defined route from the given request.
	 * @param ServerRequestInterface $request The request to match.
	 * @return RouteMatch The matched route.
	 * @throws NotFoundException
	 */
	public function match(ServerRequestInterface $request): RouteMatch
	{
		return $this->tree->match($this->splitUri($request->getUri()->getPath()), $request);
	}
}
