<?php

namespace Nest\Http\Router;

use Nest\Http\Exceptions\NotFoundException;
use Psr\Http\Message\ServerRequestInterface;

/**
 * Routes tree structure class.
 */
class RoutesTree
{
	/**
	 * Current node (static or dynamic name).
	 * @var string
	 */
	protected string $node;

	/**
	 * The defined routes for the current node.
	 * @var Route[]
	 */
	protected array $routes = [];

	/**
	 * Static children of the current node.
	 * @var array<string, RoutesTree>
	 */
	protected array $staticChildren = [];

	/**
	 * Dynamic children of the current node.
	 * @var array<string, RoutesTree>
	 */
	protected array $dynamicChildren = [];

	/**
	 * Create a new route tree structure for the given node.
	 * @param string $node Tree node name.
	 */
	public function __construct(string $node)
	{
		$this->node = $node;
	}

	/**
	 * Determine if the current node is dynamic or static.
	 * @return bool
	 */
	public function isDynamic(): bool
	{
		return $this->node[0] == ':';
	}

	/**
	 * Find the matching route for the given request.
	 * @param string[] $nodes The split URI of the route to match.
	 * @param ServerRequestInterface $request The request to match.
	 * @param array<string, string> $params URL parameters values.
	 * @return RouteMatch Matching route.
	 * @throws NotFoundException
	 */
	public function match(array $nodes, ServerRequestInterface $request, array $params = []): RouteMatch
	{
		// Get the current subnode.
		$subNode = array_shift($nodes);
		if (!empty($subNode))
			// Trim the current URI part.
			$subNode = rawurldecode(trim($subNode, "/ \n\r\t\v\0"));

		if (!empty($subNode))
		{ // There is a subnode, try to match a child.
			if (!empty($this->staticChildren[$subNode]))
			{ // Match a static child, trying to match on it.
				try
				{
					return $this->staticChildren[$subNode]->match($nodes, $request, $params);
				}
				catch (NotFoundException $notFound)
				{ // No match, trying to find another one.
				}
			}

			foreach ($this->dynamicChildren as $varName => $child)
			{ // Try to match every dynamic child.
				try
				{
					// Add subnode as route parameter.
					$params[substr($varName, 1)] = $subNode;
					return $child->match($nodes, $request, $params);
				}
				catch (NotFoundException $notFound)
				{ // No match, trying to find another one.
				}
			}

			// Still no match, throw a not found exception.
			throw new NotFoundException();
		}
		else
		{ // No subnode, the matched route can be in this node.
			foreach ($this->routes as $route)
			{ // Trying each registered route for this node.
				if ($route->match($request))
					// Found a matching route.
					return new RouteMatch($route, $params);
			}

			// No matching route found.
			throw new NotFoundException();
		}
	}

	/**
	 * Get any child with the given subnode name, if there is one.
	 * @param string $subnode Subnode name.
	 * @return RoutesTree|null The routes tree with the given subnode, or NULL if there's none.
	 */
	public function anyChild(string $subnode): ?RoutesTree
	{
		return $this->staticChildren[$subnode] ?? $this->dynamicChildren[$subnode] ?? null;
	}

	/**
	 * Register a new route for the given URI.
	 * @param string[] $nodes The split URI of the route to register.
	 * @param Route $route The route to register.
	 * @return void
	 */
	public function register(array $nodes, Route $route): void
	{
		// Get the current subnode.
		$subNode = array_shift($nodes);
		if (!empty($subNode))
			// Trim the current URI part.
			$subNode = trim($subNode, "/ \n\r\t\v\0");

		if (!empty($subNode))
		{ // There is a subnode to register, creating or getting its route tree and call register recursively.
			$child = $this->anyChild($subNode);
			if (empty($child))
			{ // No child for the current subnode, creating a new one.
				$child = new RoutesTree($subNode);
				if ($child->isDynamic())
					$this->dynamicChildren[$subNode] = $child;
				else
					$this->staticChildren[$subNode] = $child;
			}

			// Register the route in the child tree.
			$child->register($nodes, $route);
		}
		else
		{ // We reached the end of the nodes list, registering the route in this tree.
			$this->routes[] = $route;
		}
	}
}
